﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Text;

namespace ComCommonLib {

	/// <summary>
	/// Generic 2 pointer FIFO.
	/// </summary>
	/// <typeparam name="T">Element type.</typeparam>
	/// <remarks>
	/// Reasonably thread safe.  Read and write can be called from different threads.
	/// </remarks>
	public class FIFO2<T> {

		#region Buffer ---------------------------------------------------------



		/// <summary>
		/// Release both <see cref="NotEmpty"/> and <see cref="NotFull"/> prior to destroy.
		/// </summary>
		~FIFO2() {
			BufferSize = 0;
		}



		/// <summary>
		/// Set buffer size too.
		/// </summary>
		/// <param name="buffersize">Inital buffer size.</param>
		public FIFO2(int buffersize) {
			this.BufferSize= buffersize;
		}



		/// <summary>
		/// Internal buffer.
		/// </summary>
		protected T[] Buffer= new T[0];



		private int FBufferSize = 0;
		/// <summary>
		/// Size of internal buffer.
		/// </summary>
		/// <remarks>
		/// Changing it will clear the buffer contents.
		/// Setting it to 0 sets both <see cref="NotEmpty"/> and <see cref="NotFull"/>.
		/// </remarks>
		public int BufferSize {
			get { return FBufferSize; }
			set {			
				Clear();
				lock (Buffer.SyncRoot) {
					Buffer = new T[value];
					FBufferSize = value;
				}
				Clear();
			}
		}



		/// <summary>
		/// Read pointer.
		/// </summary>
		protected int ReadPtr = 0;



		/// <summary>
		/// Write pointer.
		/// </summary>
		protected int WritePtr = 0;



		/// <summary>
		/// Clear all items from internal buffer.
		/// </summary>
		public virtual void Clear() {
			lock (Buffer.SyncRoot) {
				ReadPtr = 0;
				WritePtr = 0;
				if (BufferSize == 0) {
					NotEmpty.Set();
				} else {
					NotEmpty.Reset();
				}
				NotFull.Set();
			}
		}



		/// <summary>
		/// Advance internal circular pointer.
		/// </summary>
		/// <param name="ptr"><see cref="Buffer"/> pointer.</param>
		/// <param name="add">Advance amount.</param>
		protected void AddPtr(ref int ptr, int add) {
			if (BufferSize <= add) {
				if (0 < BufferSize) {
					add %= BufferSize;
				} else {
					add = 0;
				}
			}
			ptr = ptr + add < BufferSize ? ptr + add : ptr + add - BufferSize;
		}



		/// <summary>
		/// Calculate difference between two protected circular pointers.
		/// </summary>
		/// <param name="ptr1"><see cref="Buffer"/> pointer.</param>
		/// <param name="ptr0"><see cref="Buffer"/> pointer.</param>
		/// <returns>ptr1-ptr0</returns>
		protected int SubPtr(int ptr1, int ptr0) {
			ptr1 -= ptr0;
			return 0 <= ptr1 ? ptr1 : ptr1 + BufferSize;
		}



		/// <summary>
		/// Number of used items in internal buffer.
		/// </summary>
		/// <returns>Number of used location.</returns>
		public int Used() {
			return SubPtr(WritePtr, ReadPtr);
		}



		/// <summary>
		/// Number of free items in internal buffer.
		/// </summary>
		/// <returns>Number of free items.</returns>
		public int Free() {
			return BufferSize - 1 - Used();
		}



		#endregion

		#region Signaling ------------------------------------------------------



		/// <summary>
		/// True if the internal buffer is not empty.
		/// </summary>
		/// <remarks>
		/// It is also true if <see cref="BufferSize"/>==0.
		/// </remarks>
		public ManualResetEvent NotEmpty = new ManualResetEvent(true);

	
		/// <summary>
		/// True if the internal buffer is not full.
		/// </summary>
		/// <remarks>
		/// It is also true if <see cref="BufferSize"/>==0.
		/// </remarks>
		public ManualResetEvent NotFull = new ManualResetEvent(true);



		#endregion

		#region Read -----------------------------------------------------------



		/// <summary>
		/// Get one item from internal buffer.
		/// </summary>
		/// <param name="b">Item to return.</param>
		/// <returns>True on success.</returns>
		/// <remarks>
		/// The item is consumed.
		/// </remarks>
		public virtual bool Get(out T b) {
			lock (Buffer.SyncRoot) {
				if (BufferSize == 0) {
					b = default(T);
					NotEmpty.Set();
					NotFull.Set();
					return false;
				}
				if (Used() == 0) {
					b = default(T);
					NotEmpty.Reset();
					return false;
				}
				b = Buffer[ReadPtr];
				Buffer[ReadPtr] = default(T);
				AddPtr(ref ReadPtr, 1);
				NotFull.Set();
				if (Used() == 0) {
					NotEmpty.Reset();
				}
				return true;
			}
		}



		/// <summary>
		/// Read items from internal buffer.
		/// </summary>
		/// <param name="b">Array of items to return.</param>
		/// <param name="start">Start position in array.</param>
		/// <param name="count">Max number of items requested.</param>
		/// <returns>Actual number of items returned.</returns>
		/// <remarks>
		/// Items are consumed.
		/// </remarks>
		public int Read(T[] b, int start, int count) {
			int u = SubPtr(WritePtr, ReadPtr);
			if (u < count) {
				count = u;
			}
			for (int i = 0; i < count; i++) {
				Get(out b[start + i]);
			}
			return count;
		}



		/// <summary>
		/// Read items from internal buffer.
		/// </summary>
		/// <param name="b">Array of items to return.</param>
		/// <param name="count">Max number of items requested.</param>
		/// <returns>Actual number of items returned.</returns>
		/// <remarks>
		/// Items are consumed.
		/// </remarks>
		public int Read(T[] b, int count) {
			return Read(b, 0, count);
		}



		#endregion

		#region Write ----------------------------------------------------------



		/// <summary>
		/// Put item to internal buffer.
		/// </summary>
		/// <param name="b">Item.</param>
		/// <returns>True on success.</returns>
		public virtual bool Put(T b) {
			lock (Buffer.SyncRoot) {
				if (BufferSize == 0) {
					NotEmpty.Set();
					NotFull.Set();
					return false;
				}
				if (Free() == 0) {
					NotFull.Reset();
					return false;
				}
				Buffer[WritePtr] = b;
				AddPtr(ref WritePtr, 1);
				NotEmpty.Set();
				if (Free() == 0) {
					NotFull.Reset();
				}
				return true;
			}
		}



		/// <summary>
		/// Write items to internal buffer.
		/// </summary>
		/// <param name="b">Array of items to store.</param>
		/// <param name="start">Start position in array.</param>
		/// <param name="count">Number of items in the array.</param>
		/// <returns>Actual number of items stored.</returns>
		public int Write(T[] b, int start, int count) {
			int f = Free();
			if (f < count) {
				count = f;
			}
			for (int i = 0; i < count; i++) {
				Put(b[start + i]);
			}
			return count;
		}



		/// <summary>
		/// Write items to internal buffer.
		/// </summary>
		/// <param name="b">Array of items to store.</param>
		/// <param name="count">Number of items in the array.</param>
		/// <returns>Actual number of items stored.</returns>
		public int Write(T[] b, int count) {
			return Write(b, 0, count);
		}



		#endregion

	}
}
