﻿using System;
using System.Collections.Generic;
using System.Text;

namespace ComCommonLib {

	/// <summary>
	/// Generic 3 pointer FIFO.
	/// </summary>
	/// <typeparam name="T">Item type.</typeparam>
	/// <remarks>
	/// Reasonably thread safe.  Read/peek/seek and write can be called from different threads.
	/// </remarks>
	public class FIFO3<T> : FIFO2<T> {
				
		#region Buffer ---------------------------------------------------------
		


		/// <summary>
		/// Set buffer size too.
		/// </summary>
		/// <param name="buffersize">Inital buffer size.</param>
		public FIFO3(int buffersize) : base(buffersize) {
			// Nothing
		}



		/// <summary>
		/// Seek pointer.
		/// </summary>
		protected int SeekPtr = 0;



		/// <summary>
		/// Clear all items from internal buffer.
		/// </summary>
		public override void Clear() {
			lock (Buffer.SyncRoot) {
				ReadPtr = 0;
				SeekPtr = 0;
				WritePtr = 0;
				if (BufferSize == 0) {
					NotEmpty.Set();
				} else {
					NotEmpty.Reset();
				}
				NotFull.Set();
			}
		}



		#endregion

		#region Read -----------------------------------------------------------



		/// <summary>
		/// Get one item from internal buffer.
		/// </summary>
		/// <param name="b">Item to return.</param>
		/// <returns>True on success.</returns>
		/// <remarks>
		/// The item is consumed.
		/// Seek pointer is reset.
		/// </remarks>
		public override bool Get(out T b) {
			lock (Buffer.SyncRoot) {
				if (BufferSize == 0) {
					b = default(T);
					NotEmpty.Set();
					NotFull.Set();
					return false;
				}
				if (Used() == 0) {
					b = default(T);
					NotEmpty.Reset();
					return false;
				}
				b = Buffer[ReadPtr];
				Buffer[ReadPtr] = default(T);
				AddPtr(ref ReadPtr, 1);
				NotFull.Set();
				if (Used() == 0) {
					NotEmpty.Reset();
				}
				SeekPtr = ReadPtr;
				return true;
			}
		}



		#endregion

		#region Write ----------------------------------------------------------



		/// <summary>
		/// Put item to internal buffer.
		/// </summary>
		/// <param name="b">Item.</param>
		/// <returns>True on success.</returns>
		public override bool Put(T b) {
			lock (Buffer.SyncRoot) {
				if (BufferSize == 0) {
					NotEmpty.Set();
					NotFull.Set();
					return false;
				}
				if (Free() == 0) {
					NotFull.Reset();
					return false;
				}
				Buffer[WritePtr] = b;
				AddPtr(ref WritePtr, 1);
				NotEmpty.Set();
				if (Free() == 0) {
					NotFull.Reset();
				}
				return true;
			}
		}
		


		#endregion

		#region Peek and Seek --------------------------------------------------



		/// <summary>
		/// Get one item from internal buffer.
		/// </summary>
		/// <param name="b">Item to return.</param>
		/// <returns>True on success.</returns>
		/// <remarks>
		/// The item is not consumed.
		/// </remarks>
		public bool Peek(out T b) {
			lock (Buffer.SyncRoot) {
				if (SubPtr(WritePtr, SeekPtr) <= 0) {
					b = default(T);
					return false;
				}
				b = Buffer[SeekPtr];
				AddPtr(ref SeekPtr, 1);
				return true;
			}
		}



		/// <summary>
		/// Read items from internal buffer.
		/// </summary>
		/// <param name="b">Array of items to return.</param>
		/// <param name="count">Max number of items requested.</param>
		/// <returns>Actual number of items returned.</returns>
		/// <remarks>
		/// The items are not consumed.
		/// </remarks>
		public int Peek(T[] b, int count) {
			int u = SubPtr(WritePtr, SeekPtr);
			if (u < count) {
				count = u;
			}
			for (int i = 0; i < count; i++) {
				Peek(out b[i]);
				AddPtr(ref SeekPtr, 1);
			}
			return count;
		}



		/// <summary>
		/// Consume items already peeked.
		/// </summary>
		/// <returns>Number of items freed.</returns>
		public int Discard() {
			lock (Buffer.SyncRoot) {
				int u = SubPtr(SeekPtr, ReadPtr);
				while (u-- > 0) {
					Buffer[ReadPtr] = default(T);
					AddPtr(ref ReadPtr, 1);
					NotFull.Set();
				}
				if (Used() == 0) {
					NotEmpty.Reset();
				}
				return u;
			}
		}



		/// <summary>
		/// Move the Peek pointer relative to Read pointer.
		/// </summary>
		/// <param name="n">Must be positive or zero.</param>
		/// <returns>New position.</returns>
		/// <remarks>
		/// Seek(0) allows reading the already peeked items again.
		/// </remarks>
		public int Seek(int n) {
			lock (Buffer.SyncRoot) {
				int u = SubPtr(WritePtr, ReadPtr);
				if (u < n) {
					n = u;
				}
				SeekPtr = ReadPtr;
				AddPtr(ref SeekPtr, n);
				return n;
			}
		}



		#endregion

	}
}
